A simple ETS-based key-value store for node-wide configuration.
This module provides utilities to store and retrieve data in an ETS table. It is useful for managing configuration or state that is global to a node and needs to be accessed from multiple services or processes.
This store is local to each node and is not distributed.
See the Configuration guide for more details.
Summary
Functions
Adds a value to a list stored in the store. The value is added only if it is not already present. Operation is serialized.
Deletes a value from the store.
Retrieves a value from the store. Returns default if not found.
Atomically increments or decrements a counter.
Sets a value in the store.
Updates a store key by applying a function to the current value. This operation is serialized.
Types
@type domain() :: term()
A way to separate key spaces.
Functions
Adds a value to a list stored in the store. The value is added only if it is not already present. Operation is serialized.
Deletes a value from the store.
Retrieves a value from the store. Returns default if not found.
Atomically increments or decrements a counter.
You must previously set an initial value.
Sets a value in the store.
Updates a store key by applying a function to the current value. This operation is serialized.